extends Camera2D
class_name ExtendedCamera

# TODO: look-down/up
# TODO: add documentation here!

# Camera System, General idea:
# A stack of applied cameras are kept, updated as functions are called.
# The transition is interpolated, that is most attributes
@export var camera_transition_time: float
@export var follow_camera: ExtendedCamera
@export var transition_speed_curve: Curve

@export var look_aside_enabled: bool
## The x distance which the camera looks away from the player in the direction
## the player is facing
@export
var look_aside := 0.0
## The speed in which the camera pans when the player starts facing another direction
@export
var look_aside_speed_curve: Curve

@export var inherit_position := false
@export var inherit_limit := false

@export var debug_show_origin := false

@onready var camera_stack := []

var camera_transition_time_tmp := 0.0
var camera_original_position := Vector2.ZERO

const INIT_GRACE_TRANSITION_WINDOW := 0.1
# The time node has spent processing
var time_processing := 0.0

var looking_right := true

func _ready():
	if follow_camera != null:
		camera_stack.append(follow_camera)
		await get_tree().process_frame
		reset_camera()


func _physics_process(delta: float):
	time_processing += delta
	
	if look_aside_enabled:
		var target_pos = look_aside if looking_right else - look_aside
		var curve_x = abs(target_pos - position.x) / (look_aside * 2)
		position.x += sign(target_pos) * look_aside_speed_curve.sample(
			curve_x) * delta
		position.x = clamp(position.x, -look_aside, look_aside)
		
	# Camera position update
	camera_transition_time_tmp -= delta
	if camera_transition_time_tmp <= 0.0:
		camera_transition_time_tmp = 0.0
	
	update_camera_properties()
	if debug_show_origin:
		queue_redraw()

func _draw():
	if debug_show_origin && Debug.show_gizmos:
		draw_circle(Vector2.ZERO, 10.0, Color.YELLOW)

func update_camera_properties():
	if !camera_stack.is_empty():
		# Position interpolation
		var target_pos = try_inherit("inherit_position", "global_position")
		var v = get_viewport().size
		var adjusted_target_pos = Vector2(
			clamp(target_pos.x, limit_left + v.x/2, limit_right - v.x/2),
			clamp(target_pos.y, limit_top + v.y/2, limit_bottom - v.y/2),
		)
		global_position = \
			camera_original_position \
			+ transition_speed_curve.sample(1.0 - camera_transition_time_tmp) \
			* (adjusted_target_pos - camera_original_position)
		
		# Limits
		limit_left = try_inherit("inherit_limit", "limit_left")
		limit_right = try_inherit("inherit_limit", "limit_right")
		limit_top = try_inherit("inherit_limit", "limit_top")
		limit_bottom = try_inherit("inherit_limit", "limit_bottom")

func apply_camera(c):
	if camera_stack.find(c) == -1:
		camera_stack.append(c)
		# Start transitioning to that camera
		transition(c)

func remove_camera(c):
	if camera_stack.find(c) != -1:
		# If the active camera: transition
		if camera_stack[-1] == c:
			transition(c)
		camera_stack.erase(c)

func reset_camera():
	camera_transition_time_tmp = 0.0
	update_camera_properties()
	await get_tree().create_timer(0.05).timeout
	reset_smoothing()

func transition(c):
	if time_processing >= INIT_GRACE_TRANSITION_WINDOW && c.inherit_position:
		camera_transition_time_tmp = camera_transition_time
		camera_original_position = global_position

# Recursively go through stack to inherit property
func try_inherit(enable_property, property):
	for i in camera_stack.size():
		var index = camera_stack.size() -1 -i
		var c = camera_stack[index]
		# Clean up destroyed cameras
		if c == null:
			camera_stack.remove_at(index)
		else:
			if c[enable_property] == true:
				return c[property]
	# None of the cameras had this property :(
	return self[property]
