# The global hub for all the stats.
extends CanvasLayer

@onready var control_root = $Root
@onready var categories = $Root/Stats/Categories
@onready var logs = $Root/LogScroll/Log
@onready var scroll_bar = $Root/LogScroll.get_v_scroll_bar()
@onready var log_title = $Root/LogTitle
@onready var log_count_label = $Root/LogCount

@onready var default_label = preload("res://debug/DebugLabel.tscn")

var show_gizmos = false

const ORIGINAL_SETTINGS := {
	"player_respawn" = {
		"test_1" = Vector2(20.0, 20.0)
	}
}
var settings = ORIGINAL_SETTINGS.duplicate(true)

func _ready():
	control_root.hide()
	scroll_bar.changed.connect(_on_scroll_bar_update)
	load_settings()

func _on_scroll_bar_update():
	scroll_bar.value = scroll_bar.max_value

var _log_count := 0
const LOG_COLOR_START := Color(1,1,1,0.3)
const LOG_COLOR_END := Color(0.2,0.2,0.9,0.3)
var _log_color_progress := 0.0
var _log_color_reversing := false
func put_log(value: Dictionary):
	var new_label = default_label.instantiate()
	#new_label.bbcode_enabled = true
	#new_label.fit_content = true
	new_label.text = value["description"] + " | " + value["text"]
	new_label.custom_minimum_size = Vector2(1920, 24)
	logs.add_child(new_label)
	log_title.text = "Log +"
	_log_count += 1
	
	# Change the color of the panel
	_log_color_progress += -0.3 if _log_color_reversing else 0.3
	if _log_color_progress > 1.0:
		_log_color_reversing = true
		_log_color_progress = 1.0
	if _log_color_progress < 0.0:
		_log_color_reversing = false
		_log_color_progress = 0.0
	var color = lerp(LOG_COLOR_START, LOG_COLOR_END, _log_color_progress)
	var panel := new_label.get_node("Panel")
	var style_box = StyleBoxFlat.new()
	style_box.bg_color = color
	panel.add_theme_stylebox_override("panel", style_box)
	
	log_count_label.text = str(_log_count)
	await get_tree().create_timer(0.1).timeout
	log_title.text = "Log"


func maybe_put_log(value, text: String):
	if value != null:
		value["text"] = text
		put_log(value)

# The key and string is shown like this 'key: value'
func stat_set(stat: Dictionary):
	# Check if label exists
	var found_key = false
	for child in categories.get_children():
		if child.name == stat["key"]:
			set_text(child, stat)
			found_key = true
	if not found_key:
		var new_label = default_label.instantiate()
		new_label.name = stat["key"]
		categories.add_child(new_label)
		set_text(new_label, stat)

func set_text(node, stat):
	var key = ""
	var body = ""
	var end = ""
	# Make sure key exists
	if stat.has("key"):
		key = stat["key"]
	else:
		printerr("dictionary must have key")
	# Optional body
	if stat.has("value"):
		body = stat["value"]
	# Immediate indicators
	if stat.has("indicate"):
		end = "[color=orange]*[/color] "
	node.bbcode_enabled = true
	node.fit_content = true
	node.text = key + ": " + body + " " + end
	if stat.has("indicate"):
		await get_tree().create_timer(0.5).timeout
		node.text = key + ": " + body

# Utility function for your leasure
func maybe_put_stat(stat_key, value: String):
	if stat_key != null:
		stat_key["value"] = value
		stat_set(stat_key)

func _physics_process(_delta):
	if Input.is_action_just_pressed("show_stats"):
		control_root.visible = !control_root.visible
	if Input.is_action_just_pressed("debug_show_gizmos"):
		show_gizmos = !show_gizmos
	if Input.is_action_just_pressed("debug_respawn_enemies"):
		for node in get_tree().get_nodes_in_group("debug_enemy_respawn"):
			if node.has_node("Actor"):
				var actor_node = node.get_node("Actor")
				var original_parent = node.get_parent()
				var original_position = actor_node.original_position
				var original_scene = load(actor_node.original_scene)
				if original_scene:
					node.queue_free()
					var new_node = original_scene.instantiate()
					new_node.position = original_position
					#original_parent.remove_child(actor_node)
					original_parent.add_child(new_node)
	# Add fps stat
	stat_set({ "key": "fps", "value": str(Engine.get_frames_per_second()) })
	# Clear logs on input
	if Input.is_action_just_pressed("debug_clear_log"):
		clear_log()
	
	# Debug properties of things in the world
	process_debug_property("debug_state", func(): return default_label.instantiate(), func(item): 
		item.text = item.get_parent().get_node("StateMachine").state.name
	)
	process_debug_property("debug_trail",
		func(): return preload("res://debug/Trail.tscn").instantiate(), func(_item): )

func process_debug_property(debug_name, create_func, process_func):
	for node in get_tree().get_nodes_in_group(debug_name):
		var item
		if !node.has_node(debug_name):
			item = create_func.call()
			item.name = debug_name
			node.add_child(item)
		else:
			item = node.get_node(debug_name)
		item.visible = show_gizmos
		item.process_mode = PROCESS_MODE_INHERIT if show_gizmos else PROCESS_MODE_DISABLED
		process_func.call(item)
		return item

const SETTINGS_PATH := "user://debug-settings.json"
func save_settings():
	Global.save_dict(SETTINGS_PATH, settings)
func load_settings():
	var s = Global.load_dict(SETTINGS_PATH, ORIGINAL_SETTINGS)
	if s == null:
		print("failed to load debug settings!")
		return
	else:
		settings = s

func clear_log() -> void:
	_log_count = 0
	log_count_label.text = str(_log_count)
	for child in logs.get_children():
		child.queue_free()
