extends Node
## The singleton of all singletons

@onready var pause_menu := $PauseMenu/Root
@onready var continue_button := $PauseMenu/Root/Panel/VBoxContainer/continue

const SETTINGS_PATH := "user://settings.json"

var paused := false

# Saving: settings and in-game-saves are handled differently
const ORIGINAL_SETTINGS := {
	"vsync" : true,
	"physics_interpolation" : true,
}
var settings = ORIGINAL_SETTINGS.duplicate(true)

func _ready():
	get_tree().set_auto_accept_quit(false)
	load_settings()
	pause_menu.hide()

func _process(_delta):
	if Input.is_action_just_pressed("toggle_fullscreen"):
		var new_mode
		if DisplayServer.window_get_mode() == DisplayServer.WINDOW_MODE_FULLSCREEN:
			new_mode = DisplayServer.WINDOW_MODE_WINDOWED
		else:
			new_mode = DisplayServer.WINDOW_MODE_FULLSCREEN
		DisplayServer.window_set_mode(new_mode)
	if Input.is_action_just_pressed("pause") || (paused && Input.is_action_just_pressed("ui_cancel")):
		toggle_pause()

## Saves settings
func save_settings():
	save_dict(SETTINGS_PATH, settings)

func load_settings():
	var s = load_dict(SETTINGS_PATH, ORIGINAL_SETTINGS)
	if s == null:
		print("failed to load settings!")
		return
	else:
		settings = s
		# Always apply settings, even for defaults
		for key in ORIGINAL_SETTINGS:
			self[key].call(s[key])

## Only loads fields that are in fallback
func load_dict(path: String, fallback: Dictionary):
	if FileAccess.file_exists(path):
		var settings_file = FileAccess.open(path, FileAccess.READ)
		var json_string = settings_file.get_line()
		var parse_result: Dictionary = JSON.parse_string(json_string)
		if parse_result == null:
			printerr("Failed to load settings")
			return null
		settings_file.close()
		var result = {}
		for key in fallback:
			if parse_result.has(key):
				result[key] = parse_result[key]
		return result
	else:
		return null

func save_dict(path: String, dict: Dictionary):
	var json_string = JSON.stringify(dict)
	var settings_file = FileAccess.open(path, FileAccess.WRITE)
	settings_file.store_string(json_string)
	settings_file.close()

func toggle_pause():
	paused = !paused
	pause_menu.visible = paused
	get_tree().paused = paused
	if paused:
		continue_button.grab_focus()


func _on_exit_game_pressed():
	get_tree().quit()


func vsync(toggled_on: bool) -> void:
	settings["vsync"] = toggled_on
	DisplayServer.window_set_vsync_mode(DisplayServer.VSYNC_ENABLED if toggled_on else DisplayServer.VSYNC_DISABLED)
	save_settings()


func physics_interpolation(toggled_on: bool) -> void:
	settings["physics_interpolation"] = toggled_on
	get_tree().physics_interpolation = toggled_on
	save_settings()

func _notification(what):
	if what == NOTIFICATION_WM_CLOSE_REQUEST:
		print('by by')
		get_tree().quit()
